#include <Eet.h>
#include <Evas.h>
#include <Ecore_File.h>
#include "points.h"
#include "eli_hiscore.h"
#include <stdio.h>
#include <string.h>

/* avoid warings */
typedef void * (*list_next)    (void *);
typedef void * (*list_append)  (void *, void *);
typedef void * (*list_data)    (void *);
typedef void * (*list_free)    (void *);
typedef void   (*hash_foreach) (void *,
                               int (*func) (void *, const char *, void *, void *), 
                               void *);
typedef void * (*hash_add)     (void *, const char *, void *);
typedef void   (*hash_free)    (void *);

/* globals */
static Eet_Data_Descriptor * edd_hiscore;
static Eet_Data_Descriptor * edd_entry;
static char * eet_file_name;

/* internals declaration */
static int                   _eli_highscore_list_sort_bad  (void * ent1, void * ent2);
static int                   _eli_highscore_list_sort_good (void * ent1, void * ent2);
static Eli_Highscore       * _eli_highscore_new            (void);
static Evas_List           * _eli_entries_list_copy        (Evas_List * l);
static Eli_Highscore_Entry * _eli_entry_new                (const char * username, float points);
static float                 _min                          (float v1, float v2);
static float                 _max                          (float v1, float v2);

/* ***************************************************************************
 *     Externals
 * ***************************************************************************/

void eli_highscore_init(const char * file)
{
    eet_init();

    edd_entry = eet_data_descriptor_new("Eli_Highscore_Entry",
                                        sizeof(Eli_Highscore_Entry),
                                        (list_next) evas_list_next,
                                        (list_append) evas_list_append,
                                        (list_data) evas_list_data,
                                        (list_free) evas_list_free,
                                        (hash_foreach) evas_hash_foreach,
                                        (hash_add) evas_hash_add,
                                        (hash_free) evas_hash_free);

    EET_DATA_DESCRIPTOR_ADD_BASIC(edd_entry, Eli_Highscore_Entry,
                                  "username", username, EET_T_STRING);
    EET_DATA_DESCRIPTOR_ADD_BASIC(edd_entry, Eli_Highscore_Entry,
                                  "points", points, EET_T_FLOAT);

    edd_hiscore = eet_data_descriptor_new("Eli_Highscore",
                                          sizeof(Eli_Highscore),
                                          (list_next) evas_list_next,
                                          (list_append) evas_list_append,
                                          (list_data) evas_list_data,
                                          (list_free) evas_list_free,
                                          (hash_foreach) evas_hash_foreach,
                                          (hash_add) evas_hash_add,
                                          (hash_free) evas_hash_free);

    EET_DATA_DESCRIPTOR_ADD_LIST(edd_hiscore, Eli_Highscore,
                                 "entries", entries, edd_entry);
    EET_DATA_DESCRIPTOR_ADD_BASIC(edd_hiscore, Eli_Highscore,
                                  "type", type, EET_T_INT);
    eet_file_name = strdup(file);
}

void eli_highscore_shutdown(void)
{
    eet_data_descriptor_free(edd_hiscore);
    eet_data_descriptor_free(edd_entry);
    free(eet_file_name);
    eet_shutdown();
}

Evas_Bool eli_highscore_entry_add(const char * game, const char * username,
                                  float points, pointsType type)
{
    Evas_List * l = NULL;
    Eli_Highscore_Entry * entry;
    Eli_Highscore hiscore;
    Eet_File * file;
    int (*list_sort) (void *, void *);
    int count = 0;

    entry = _eli_entry_new(username, points);

    l = eli_highscore_get(game);
    if (l) count = evas_list_count(l);
    else count = 0;

    /* select the right sorting function */
    switch (type) {
    case POINTS_TYPE_INTEGER_BAD:
    case POINTS_TYPE_FLOAT_BAD:
        list_sort = _eli_highscore_list_sort_bad;
        break;
    default:
        list_sort = _eli_highscore_list_sort_good;
        break;
    }

    /* 10 entries should be enough */
    if (count >= 10) {
        Evas_List * last_l;
        Eli_Highscore_Entry * last_e;

        last_l = evas_list_last(l);
        last_e = evas_list_data(last_l);

        if ((*list_sort) (last_e, entry) < 0) {
            free(entry->username);
            free(entry);
            eli_highscore_free(l);
            return 0;
        }
    }

    l = evas_list_append(l, entry);
    if (count) l = evas_list_sort(l, (count + 1), list_sort);
    if (count >= 10) l = evas_list_remove_list(l, evas_list_last(l));

    if (ecore_file_exists(eet_file_name)) {
        file = eet_open(eet_file_name, EET_FILE_MODE_READ_WRITE);
    }
    else {
        file = eet_open(eet_file_name, EET_FILE_MODE_WRITE);
        if (!file) fprintf(stderr, "Could not create file\n!");
    }

    hiscore.entries = l;
    hiscore.type = type;

    eet_data_write(file, edd_hiscore, game, &hiscore, 1);
    eet_close(file);

    eli_highscore_free(l);

    return 1;
}

Evas_Bool eli_highscore_accept(const char * game, float points,
                               pointsType type)
{
    Evas_List * l, * list;
    float m;
    float (*select) (float, float);

    m = 0.0;
    list = eli_highscore_get(game);

    if (!list) return 1 == 1;
    if (evas_list_count(list) < 10) return 1 == 1;

    /* select the right min/max function */
    switch (type) {
    case POINTS_TYPE_INTEGER_BAD:
    case POINTS_TYPE_FLOAT_BAD:
        select = _max;
        break;
    default:
        select = _min;
        break;
    }

    for (l = list; l; l = l->next) {
        Eli_Highscore_Entry * e;

        e = (Eli_Highscore_Entry *) evas_list_data(l);
        m = (*select) (m, e->points);
    }

    eli_highscore_free(list);

    return m == select(m, points);
}

Evas_List * eli_highscore_get(const char * game)
{
    Eet_File * file = NULL;
    Eli_Highscore * hiscore = NULL;
    Evas_List * l = NULL;

    if (!ecore_file_exists(eet_file_name)) return NULL;
    file = eet_open(eet_file_name, EET_FILE_MODE_READ);
    if (!file) return NULL;

    hiscore = (Eli_Highscore *) eet_data_read(file, edd_hiscore, (char*) game);

    if (hiscore) l = _eli_entries_list_copy(hiscore->entries);
    eet_close(file);

    if (hiscore) free(hiscore);

    return l;
}

pointsType eli_highscore_points_type_get(const char * game)
{
    Eet_File * file = NULL;
    Eli_Highscore * hiscore = NULL;
    pointsType type = POINTS_TYPE_UNKNOWN;

    if (!ecore_file_exists(eet_file_name)) return POINTS_TYPE_UNKNOWN;
    
    file = eet_open(eet_file_name, EET_FILE_MODE_READ);
    if (!file) return POINTS_TYPE_UNKNOWN;

    hiscore = (Eli_Highscore *) eet_data_read(file, edd_hiscore, (char*) game);

    if (hiscore) type = hiscore->type;
    eet_close(file);

    if (hiscore) free(hiscore);
    return type;
}

void eli_highscore_free(Evas_List * list)
{
    Evas_List * l;

    if (!list) return;
    l = list;

    while (l) {
        Eli_Highscore_Entry * entry;

        entry = (Eli_Highscore_Entry *) evas_list_data(l);

        if (entry)
            if (entry->username) {
                free(entry->username);
                entry->username = NULL;
            }

        list = evas_list_remove_list(list, l);
        l = l->next;
    }
}

/* ***************************************************************************
 *     Internals							               
 * ***************************************************************************/

static int _eli_highscore_list_sort_bad(void * ent1, void * ent2)
{
    Eli_Highscore_Entry * entry1, * entry2;

    entry1 = (Eli_Highscore_Entry *) ent1;
    entry2 = (Eli_Highscore_Entry *) ent2;

    return (entry1->points - entry2->points) * 100.0;
}

static int _eli_highscore_list_sort_good(void * ent1, void * ent2)
{
    Eli_Highscore_Entry * entry1, * entry2;

    entry1 = (Eli_Highscore_Entry *) ent1;
    entry2 = (Eli_Highscore_Entry *) ent2;

    return (entry2->points - entry1->points) * 100.0;
}

static Eli_Highscore * _eli_highscore_new()
{
    Eli_Highscore * eh;

    eh = (Eli_Highscore *) malloc(sizeof(Eli_Highscore));
    eh->entries = NULL;

    return eh;
}

static Evas_List * _eli_entries_list_copy(Evas_List * l)
{
    Evas_List * new_list = NULL;
    Evas_List * list;

    for (list = l; list; list = list->next) {
        Eli_Highscore_Entry * entry, * old_e;

        old_e = (Eli_Highscore_Entry *) evas_list_data(list);
        entry = _eli_entry_new(old_e->username, old_e->points);

        new_list = evas_list_append(new_list, entry);
    }

    return new_list;
}

static Eli_Highscore_Entry * _eli_entry_new(const char * username, float points)
{
    Eli_Highscore_Entry * entry;

    entry = (Eli_Highscore_Entry *) malloc(sizeof(Eli_Highscore_Entry));
    entry->points = points;
    entry->username = strdup(username);

    return entry;
}

static float _min(float v1, float v2)
{
    return (v1 < v2) ? v1 : v2;
}

static float _max(float v1, float v2)
{
    return (v1 > v2) ? v1 : v2;
}
