#include "Eli_App.h"
#include "eli_statistics.h"

/* internal declaration */
static void       _win_game_cb             (int end, float points, pointsType ptype, void * data);
static void       _points_cb               (float points, pointsType ptype, void * data);
static void       _eli_app_game_new_force  (Eli_App * eap);
static int        _eli_app_wait_for_end    (void * data);

/* 
 *   only used in _win_game_cb and ...
 */
struct _Eli_App_End
{
    Eli_App  * eap;
    int        end;
    pointsType type;
    float      points;
};

/* 
 *   external functions
 */

void eli_app_menu_container_fill(Eli_App * eap, Evas_Object * container)
{
    const char ** games;
    char * file;

    file = ecore_config_theme_with_path_from_name_get(eap->theme.gui.current);
    games = elitaire_available_games_get();

    for (int i = 0; games[i][0]; i++) {
        Evas_Object * item;

        item = edje_object_add(eap->evas);

        if (edje_object_file_set(item, file, "element")) {
            Evas_Coord w, h;

            edje_object_size_min_get(item, NULL, &h);
            evas_object_geometry_get(container, NULL, NULL, &w, NULL);
            evas_object_resize(item, w, h);

            if (edje_object_part_exists(item, "elitaire_element_value")) {
                edje_object_part_text_set(item, "elitaire_element_value",
                                          games[i]);
                evas_object_show(item);

                edje_object_signal_callback_add(item, "item_selected",
                                                "item_selected",
                                                _eli_app_new_game_cb, eap);

                esmart_container_element_append(container, item);
            }
            else {
                fprintf(stderr, _("Elitaire Error in %s: "
                                  "missing element_value part!\n"), file);
                evas_object_del(item);
            }
        }
        else {
            fprintf(stderr,
                    _("Elitaire Error in %s: missing element group!\n"),
                    file);
            evas_object_del(item);
        }
    }
}

void eli_app_game_end(Eli_App * eap)
{
    if (eap->elitaire) {
        evas_object_del(eap->elitaire);
        eap->elitaire = NULL;
    }
    
    edje_object_signal_emit(eap->gui, "win,off", "");
    
    if (eap->next.game) {
        if (!strcmp(eap->next.game, "exit")) ecore_main_loop_quit();
        else _eli_app_game_new_force(eap);
    }
}

void eli_app_game_new(Eli_App * eap, const char * game)
{
    eap->next.game = strdup(game);
    /* end current game */
    if (eap->elitaire) {
        elitaire_object_giveup(eap->elitaire);
    }
    else {
        _eli_app_game_new_force(eap);
    }
}

void eli_app_velocity_set(Eli_App * eap, int velocity)
{
    double v;
    Evas_Object * obj;

    v = eli_app_velocity_real_to_percent(velocity);

    if ((obj = edje_object_part_swallow_get(eap->gui, "elitaire_menu_bar"))) {

        edje_object_part_drag_value_set(obj, "elitaire_velocity_slider", v,
                                        0);

        if (edje_object_part_exists(obj, "elitaire_velocity_value")) {
            char buffer[5];

            snprintf(buffer, 5, "%i", velocity);
            edje_object_part_text_set(obj, "elitaire_velocity_value", buffer);
        }
    }
}

int eli_app_velocity_get(Eli_App * eap)
{
    Evas_Object * obj;

    if ((obj = edje_object_part_swallow_get(eap->gui, "elitaire_menu_bar"))) {
        double v;
        
        edje_object_part_drag_value_get(obj, "elitaire_velocity_slider", &v,
                                        NULL);
        
        return eli_app_velocity_percent_to_real(v);
    }
    
    return eli_app_velocity_percent_to_real(0);
}

/*  
 *  internal functions
 */


/* to win not window */
static void _win_game_cb(int end, float points, pointsType ptype, void * data)
{
    Eli_App * eap;

    eap = (Eli_App *) data;
    
    if (eap) {
        _Eli_App_End * eend = NULL;
        char * game;

        game = eap->current.game;

        elitaire_object_pause(eap->elitaire);
        if (end == 0) {
            if (edje_object_part_exists(eap->gui, "elitaire_win_text")) {
                edje_object_part_text_set(eap->gui, "elitaire_win_text",
                                          _("You win!"));
            }
            eli_statistics_win_add(game);
        }
        else if (end == 1) {
            if (edje_object_part_exists(eap->gui, "elitaire_win_text")) {
                edje_object_part_text_set(eap->gui, "elitaire_win_text",
                                          _("You lose!"));
            }
            eli_statistics_lost_add(game);
            eli_app_game_end(eap);
            return;
        }
        else {
            if (edje_object_part_exists(eap->gui, "elitaire_win_text")) {
                edje_object_part_text_set(eap->gui, "elitaire_win_text",
                                          _("Game Over"));
            }
            eli_statistics_lost_add(eap->current.game);
        }
        edje_object_signal_emit(eap->gui, "win,on", "");

        eend = (_Eli_App_End *) malloc(sizeof(_Eli_App_End));
        eend->eap = eap;
        eend->end = end;
        eend->points = points;
        eend->type = ptype;

        elitaire_object_wait_for_end_of_jobs(eap->elitaire, _eli_app_wait_for_end,
                                      eend);
    }
}

static int _eli_app_wait_for_end(void * data)
{
    _Eli_App_End * eend;

    eend = (_Eli_App_End *) data;
    if (eend->end == 0) {
        if (eli_highscore_accept
            (eend->eap->current.game, eend->points, eend->type)) {
            win_highscore_add(eend->eap, eend->points, eend->type);
        }
        else {
            eli_app_game_end(eend->eap);
        }
    }
    else if (eend->end == 1) {
        fprintf(stderr,
                "that shouldn't happen\n void _eli_app_wait_for_end\n");
    }
    else {
        if (eli_highscore_accept
            (eend->eap->current.game, eend->points, eend->type))
            win_highscore_add(eend->eap, eend->points, eend->type);
        else
            eli_app_game_end(eend->eap);
    }
    free(eend);

    return 0;
}

static void _points_cb(float points, pointsType ptype, void *data)
{
    Eli_App * eap;

    eap = (Eli_App *) data;
    if (eap) {
        if (edje_object_part_exists(eap->gui, "elitaire_points")) {
            char * pstring;

            pstring = pointsType_point_string_get(points, ptype);
            edje_object_part_text_set(eap->gui, "elitaire_points", pstring);
            free(pstring);
        }
    }
}

static void _eli_app_game_new_force(Eli_App * eap)
{
    Evas_Coord x, y, w, h;
    Evas_Coord l, r, t, b;
    Evas_Coord min_w, min_h;
    char * card_theme;
    int vel;
    Evas_Bool bo;

    if (eap->current.game) {
        free(eap->current.game);
        eap->current.game = NULL;
    }
    /* OK now the we can make the new game to current game */
    eap->current.game = eap->next.game;
    eap->next.game = NULL;
    edje_object_signal_emit(eap->gui, "win,off", "");

    /* init elitaire, set the theme, ... */
    if (!(eap->elitaire = elitaire_object_new(eap->evas, eap->current.game)))
        return;
    card_theme =
        ecore_config_theme_with_path_from_name_get(eap->theme.cards.current);
    /* get and set the offset */
    eli_app_theme_offset_get(eap, &l, &r, &t, &b);
    elitaire_object_offset_set(eap->elitaire, l, r, t, b);

    if (elitaire_object_file_set(eap->elitaire, card_theme)) {
        elitaire_object_size_min_get(eap->elitaire, &w, &h);
    }
    else {
        fprintf(stderr, _("Elitaire Error: cards file %s not found!\n"),
                eap->theme.cards.current);
        return;
    }
    free(card_theme);
    card_theme = NULL;

    /* find and set the minimum size */
    elitaire_object_size_min_get(eap->elitaire, &min_w, &min_h);
    if (!edje_object_part_exists(eap->gui, "elitaire_card_field")) {
        fprintf(stderr,
                _("Elitaire Error in %s: elitaire_card_field does not exist!\n"),
                eap->theme.gui.current);
        return;
    }
    edje_object_part_geometry_get(eap->gui, "elitaire_card_field", &x, &y, &w, &h);
    
    if (w < min_w || h < min_h) {
        Evas_Coord win_h, win_w;

        ecore_evas_geometry_get(eap->ee, NULL, NULL, &win_w, &win_h);
        ecore_evas_resize(eap->ee, win_w + min_w - w, win_h + min_h - h);
    }
    /* move and resize */
    evas_object_move(eap->elitaire, x, y);
    evas_object_resize(eap->elitaire, w, h);
    evas_object_show(eap->elitaire);

    /* tell elitaire the ecore_config vlaues */
    vel = ecore_config_int_get("velocity");
    elitaire_object_velocity_set(eap->elitaire, vel);

    vel = ecore_config_int_get("frame_rate");
    elitaire_object_frame_rate_set(eap->elitaire, vel);

    vel = ecore_config_int_get("lazy");
    elitaire_object_lazy_mode_set(eap->elitaire, vel);

    bo = ecore_config_boolean_get("/graphic/animations");
    elitaire_object_animations_set(eap->elitaire, bo);

    bo = ecore_config_boolean_get("/graphic/shadows");
    elitaire_object_shadows_set(eap->elitaire, bo);

    /* swallow the elitaire object */
    edje_object_part_swallow(eap->gui, "elitaire_card_field", eap->elitaire);
    /* Add the callbacks */
    elitaire_object_callback_win_add(eap->elitaire, _win_game_cb, eap);
    elitaire_object_callback_points_add(eap->elitaire, _points_cb, eap);
    
    /* and now deal */
    elitaire_object_deal(eap->elitaire);
    return;
}

void _eli_app_new_game_cb(void * data, Evas_Object * o, const char * emission,
                          const char * source)
{
    Eli_App * eap;

    eap = (Eli_App *) data;

    if (!strcmp(source, "")) {
        if (eap->current.game) {
            eli_app_game_new(eap, eap->current.game);
        }
        else return;
    }
    else if (strcmp(source, "item_selected")) eli_app_game_new(eap, emission);
    else {
        const char * game;

        game = edje_object_part_text_get(o, "elitaire_element_value");
        eli_app_game_new(eap, game);
    }
}
